
-- Creation des tables et contraintes

CREATE TABLE Agence
 (numAg VARCHAR2(6), nomAg VARCHAR(20),
  CONSTRAINT pk_Ag PRIMARY KEY(numAg));

CREATE TABLE Employes
 (numEmp VARCHAR2(5), nomEmp VARCHAR(20), age NUMBER(2), resp VARCHAR2(6),
 CONSTRAINT fk_EmpAg FOREIGN KEY(resp) REFERENCES Agence(numAg),
  CONSTRAINT pk_Emp PRIMARY KEY(numEmp));

CREATE TABLE Travaille
 (numAg VARCHAR2(6), numEmp VARCHAR2(5), dateDeb DATE, dateFin DATE,
 CONSTRAINT pk_Trav PRIMARY KEY(numAg, numEmp, dateDeb),
 CONSTRAINT fk_TravAg FOREIGN KEY(numAg) REFERENCES Agence(numAg),
 CONSTRAINT fk_TravEmp FOREIGN KEY(numEmp) REFERENCES Employes(numEmp));

CREATE TABLE ContratsVendus
 (numAg VARCHAR2(6), numEmp VARCHAR2(5), numCli NUMBER(5), nomProd VARCHAR2(10), dateCont DATE,
 CONSTRAINT pk_Cont PRIMARY KEY(numAg, numEmp,numCli, nomProd),
 CONSTRAINT fk_ContAg FOREIGN KEY(numAg) REFERENCES Agence(numAg),
 CONSTRAINT fk_ContEmp FOREIGN KEY(numEmp) REFERENCES Employes(numEmp),
 CONSTRAINT fk_ContCli FOREIGN KEY(numCli) REFERENCES Client(num));


INSERT INTO Agence VALUES ('AG1','St-Alban');
INSERT INTO Agence VALUES ('AG2','St-Sulpice');
INSERT INTO Agence VALUES ('AG3','Paris 6');

--
INSERT INTO Employes VALUES ('E1','Lamothe P.', 37, NULL);
INSERT INTO Employes VALUES ('E2','Sry P.', 39, 'AG2');
INSERT INTO Employes VALUES ('E3','Laborde G.', 43, NULL);
INSERT INTO Employes VALUES ('E4','Roche C.', 40, 'AG3');
INSERT INTO Employes VALUES ('E5','Prissel F.', 36,'AG1');

--

INSERT INTO Travaille VALUES ('AG2','E1','01-01-2000', '31-12-2002');
INSERT INTO Travaille VALUES ('AG1','E1','01-01-2003', NULL);
INSERT INTO Travaille VALUES ('AG3','E2','01-01-1998', '01-12-2001');
INSERT INTO Travaille VALUES ('AG1','E2','02-12-2001', '31-12-2001');
INSERT INTO Travaille VALUES ('AG2','E2','01-01-2002', NULL);
INSERT INTO Travaille VALUES ('AG1','E3','01-01-1995', '31-12-2000');
INSERT INTO Travaille VALUES ('AG3','E3','01-01-2001', '31-12-2003');
INSERT INTO Travaille VALUES ('AG2','E3','01-01-2004', NULL);
INSERT INTO Travaille VALUES ('AG3','E4','01-01-2004', NULL);
INSERT INTO Travaille VALUES ('AG1','E5','01-01-2004', NULL);

--

INSERT INTO ContratsVendus VALUES ('AG2','E1',1, 'PEL', '02-01-2000');
INSERT INTO ContratsVendus VALUES ('AG2','E1',2, 'CEL', '03-01-2000');
INSERT INTO ContratsVendus VALUES ('AG2','E1',2, 'PEL', '03-01-2000');
INSERT INTO ContratsVendus VALUES ('AG1','E1',3, 'CODEVI', '02-01-2003');
INSERT INTO ContratsVendus VALUES ('AG1','E1',4, 'CEL', '04-01-2003');

INSERT INTO ContratsVendus VALUES ('AG3','E2',5, 'CODEVI', '04-01-1998');
INSERT INTO ContratsVendus VALUES ('AG3','E2',5, 'LIVRET-A', '04-01-1998');
INSERT INTO ContratsVendus VALUES ('AG1','E2',4, 'LIVRET-A', '24-12-2001');
INSERT INTO ContratsVendus VALUES ('AG2','E2',4, 'LIVRET-A', '04-01-2002');

INSERT INTO ContratsVendus VALUES ('AG1','E3',4, 'CODEVI', '24-12-1997');
INSERT INTO ContratsVendus VALUES ('AG1','E3',4, 'SICAV', '24-12-1997');

--

CREATE TYPE employes_type AS OBJECT
 (numEmp VARCHAR2(5), nomEmp VARCHAR(20), age NUMBER(2), resp VARCHAR2(6))
/
CREATE TYPE elt_nt_equipe_type AS OBJECT
 (refEmp REF employes_type, dateEnt DATE)
/
CREATE TYPE equipe_nt_type AS TABLE OF elt_nt_equipe_type
/
CREATE TYPE elt_nt_contrats_type AS OBJECT
 (refEmp REF employes_type, refCli REF client_type, nomProd VARCHAR2(10), dateCont DATE)
/
CREATE TYPE contrats_nt_type AS TABLE OF elt_nt_contrats_type
/
CREATE TYPE agenceVOR_type AS OBJECT
 (numAg VARCHAR2(6), nomAg VARCHAR(20), equipe_nt equipe_nt_type, contrats_nt contrats_nt_type)
/

CREATE VIEW Employes_VOR OF employes_type 
 WITH OBJECT IDENTIFIER(numEmp) AS SELECT * FROM Employes;

CREATE VIEW Client_VOR OF client_type 
 WITH OBJECT IDENTIFIER(num) AS SELECT * FROM Client;

CREATE OR REPLACE VIEW Agence_VOR OF agenceVOR_type
 WITH OBJECT IDENTIFIER(numAg)
 AS SELECT a.numAg, a.nomAg,
           CAST( MULTISET 
                (SELECT REF(e), t.dateDeb
                        FROM  Employes_VOR e, Travaille t
                        WHERE t.numEmp = e.numEmp 
                        AND   t.dateFin IS NULL 
		        AND   t.numAg  = a.numAg) AS equipe_nt_type
               ) AS equipeCAST, 
           CAST( MULTISET 
                (SELECT MAKE_REF(Employes_VOR, cv.numEmp), MAKE_REF(Client_VOR, cv.numCli), cv.nomProd, cv.dateCont
                        FROM ContratsVendus cv
                        WHERE cv.numAg = a.numAg) AS contrats_nt_type
               ) AS contratsCAST
  FROM Agence a;

SELECT * FROM Agence_VOR;

--40
SELECT  a.numAg, a.nomAg, ant.refEmp.nomEmp, ant.refEmp.age, ant.dateEnt
 FROM Agence_VOR a, TABLE(a.equipe_nt) ant 
 ORDER BY 1,4,5;

--41
-- Liste des contrats vendus (nom du produit et date de la vente) 
-- classe par agence et par numro de client.

SELECT  ant.nomProd, a.numAg, ant.refEmp.nomEmp, ant.refCli.nom, ant.dateCont
 FROM Agence_VOR a, TABLE(a.contrats_nt) ant 
 ORDER BY 1,2;

--42.	Dtails des contrats vendus par le responsable de lagence classe par agence.

SELECT  ant.nomProd, a.numAg, ant.refEmp.nomEmp, ant.refCli.nom, ant.dateCont
 FROM Agence_VOR a, TABLE(a.contrats_nt) ant, Employes e
 WHERE e.resp = a.numAg
 AND   ant.refEmp.numEmp = e.numEmp
 ORDER BY 1,2;

--43


ALTER TYPE agenceVOR_type 
 ADD MEMBER PROCEDURE passeContrat(nEmp IN VARCHAR2, nCli IN NUMBER, nomProduit IN VARCHAR2)
 CASCADE;

ALTER TYPE agenceVOR_type 
 ADD MEMBER PROCEDURE mutation(nEmp IN VARCHAR2, nAgCible IN VARCHAR2)
 CASCADE;


CREATE OR REPLACE TYPE BODY agenceVOR_type AS 
  MEMBER PROCEDURE passeContrat(nEmp IN VARCHAR2, nCli IN NUMBER, nomProduit IN VARCHAR2) IS
    trace    NUMBER(1) := 1 ;
    v_nom    VARCHAR2(30);
    v_nombre NUMBER(4);
  BEGIN
	SELECT nom INTO v_nom FROM Client WHERE num = nCli;
        trace := 2;
	SELECT nomEmp INTO v_nom FROM Employes WHERE numEmp = nEmp;

        SELECT COUNT(*) INTO v_nombre FROM ContratsVendus
           WHERE numAg = SELF.numAg
	   AND  numEmp = nEmp 
           AND  numCli = nCli 
	   AND nomProd = nomProduit;

	IF (v_nombre <> 0) THEN 
	  DBMS_OUTPUT.PUT_LINE('Produit dj vendu par cet employ  ce client');
	ELSE	
	  INSERT INTO ContratsVendus VALUES(SELF.numAg,nEmp,nCli,nomProduit,SYSDATE);
	  DBMS_OUTPUT.PUT_LINE('Vente enregistre.');
	  COMMIT;
	END IF;
  EXCEPTION
    WHEN NO_DATA_FOUND THEN
     IF trace = 1 THEN
       DBMS_OUTPUT.PUT_LINE('Pas de client de ce numro : ' || nCli);
     ELSE
       DBMS_OUTPUT.PUT_LINE('Pas d''employ de ce numro : ' || nEmp);
     END IF;
  WHEN OTHERS THEN
       DBMS_OUTPUT.PUT_LINE('Erreur d''Oracle ' || SQLERRM  || '(' || SQLCODE || ')');
  END passeContrat;
--  
 MEMBER PROCEDURE  mutation(nEmp IN VARCHAR2, nAgCible IN VARCHAR2) IS
   trace    NUMBER(1) := 1 ;
   v_nom    VARCHAR2(30);
   v_nombre NUMBER(4);
 BEGIN
  SELECT nomAg INTO v_nom FROM Agence WHERE numAg = nAgCible ;
  trace := 2;
  SELECT nomEmp INTO v_nom FROM Employes WHERE numEmp = nEmp;
  trace := 3;
  SELECT numEmp INTO v_nom FROM Travaille 
         WHERE numEmp =  nEmp 
         AND numAg    =  SELF.numAg 
         AND dateFin  IS NULL;

   SELECT  COUNT(*) INTO v_nombre FROM Employes WHERE resp = SELF.numAg AND numEmp = nEmp;
   IF (v_nombre <> 0) THEN 
	  DBMS_OUTPUT.PUT_LINE('Il faut nommer un nouveau responsable pour l''agence '|| SELF.numAg);
   END IF;
   UPDATE Travaille 
         SET   dateFin = SYSDATE
         WHERE numAg   = SELF.numAg 
         AND   numEmp  = nEmp;
   UPDATE Employes
         SET   resp   = NULL
         WHERE numEmp = nEmp;
   INSERT INTO Travaille VALUES(nAgCible, nEmp, SYSDATE, NULL);
   DBMS_OUTPUT.PUT_LINE('Mutation de ' || SELF.numAg || '  ' || nAgCible || ' enregistre.');
   COMMIT;
 
 EXCEPTION
    WHEN NO_DATA_FOUND THEN
     IF trace = 1 THEN
       DBMS_OUTPUT.PUT_LINE('Pas d''agence de ce numro : ' || nAgCible);
     ELSIF trace = 2 THEN
       DBMS_OUTPUT.PUT_LINE('Pas d''employ de ce numro : ' || nEmp);
     ELSIF trace = 3 THEN
       DBMS_OUTPUT.PUT_LINE('L''employ n''appartient pas  l''agence ' || SELF.numAg);
     END IF;
  WHEN OTHERS THEN
       DBMS_OUTPUT.PUT_LINE('Erreur d''Oracle ' || SQLERRM  || '(' || SQLCODE || ')');
  END mutation;
END;
/
show err

--Tests 43
SELECT * FROM ContratsVendus;

SELECT  ant.nomProd, a.numAg, ant.refEmp.nomEmp, ant.refCli.nom, ant.dateCont
 FROM Agence_VOR a, TABLE(a.contrats_nt) ant 
 ORDER BY 1,2;

SET SERVEROUT ON
DECLARE
  une_Agence agenceVOR_type;
BEGIN
  SELECT VALUE(ag) INTO une_Agence 
	FROM Agence_VOR ag WHERE ag.numAg = 'AG1';
  une_Agence.passeContrat('E2', 4, 'PEL');
END;
/

SELECT  ant.nomProd, a.numAg, ant.refEmp.nomEmp, ant.refCli.nom, ant.dateCont
 FROM Agence_VOR a, TABLE(a.contrats_nt) ant 
 ORDER BY 1,2;

SELECT * FROM ContratsVendus;

--Test erreur dj vendu
DECLARE
  une_Agence agenceVOR_type;
BEGIN
  SELECT VALUE(ag) INTO une_Agence 
	FROM Agence_VOR ag WHERE ag.numAg = 'AG1';
  une_Agence.passeContrat('E2', 4, 'PEL');
END;
/

--emp inexistant
DECLARE
  une_Agence agenceVOR_type;
BEGIN
  SELECT VALUE(ag) INTO une_Agence 
	FROM Agence_VOR ag WHERE ag.numAg = 'AG1';
  une_Agence.passeContrat('E9', 4, 'PEL');
END;
/

--cli inexistant
DECLARE
  une_Agence agenceVOR_type;
BEGIN
  SELECT VALUE(ag) INTO une_Agence 
	FROM Agence_VOR ag WHERE ag.numAg = 'AG1';
  une_Agence.passeContrat('E2', 0, 'PEL');
END;
/

--44

--Tests 44

SELECT * FROM Travaille ORDER BY 1;

SET SERVEROUT ON
DECLARE
  une_Agence agenceVOR_type;
BEGIN
  SELECT VALUE(ag) INTO une_Agence 
	FROM Agence_VOR ag WHERE ag.numAg = 'AG1';
  une_Agence.mutation('E1', 'AG2');
END;
/

SELECT * FROM Travaille ORDER BY 1;

--Tests erreur

--ag inexistante

SET SERVEROUT ON
DECLARE
  une_Agence agenceVOR_type;
BEGIN
  SELECT VALUE(ag) INTO une_Agence 
	FROM Agence_VOR ag WHERE ag.numAg = 'AG1';
  une_Agence.mutation('E1', 'AG9');
END;
/

--cli inexistant
SET SERVEROUT ON
DECLARE
  une_Agence agenceVOR_type;
BEGIN
  SELECT VALUE(ag) INTO une_Agence 
	FROM Agence_VOR ag WHERE ag.numAg = 'AG1';
  une_Agence.mutation('E9', 'AG2');
END;
/

--emp resp de l'agence source
SET SERVEROUT ON
DECLARE
  une_Agence agenceVOR_type;
BEGIN
  SELECT VALUE(ag) INTO une_Agence 
	FROM Agence_VOR ag WHERE ag.numAg = 'AG1';
  une_Agence.mutation('E5', 'AG2');
END;
/

--emp inexistant dans l'agence source
SET SERVEROUT ON
DECLARE
  une_Agence agenceVOR_type;
BEGIN
  SELECT VALUE(ag) INTO une_Agence 
	FROM Agence_VOR ag WHERE ag.numAg = 'AG1';
  une_Agence.mutation('E5', 'AG2');
END;
/

